from typing import List

from guet.commands.command import Command
from guet.config.get_config import get_config
from guet.config.set_config import set_config
from guet.settings.settings import Settings


class ConfigSetCommand(Command):
    HELP_MESSAGE = 'usage: guet config [--<key>=<value> ...]'
    SHORT_HELP_MESSAGE = 'Change setting values'

    def execute_hook(self) -> None:
        config = get_config()
        key_and_value = self._separate_key_and_value()
        self._append_key_and_value_to_config(config, key_and_value)
        set_config(config)

    def help(self) -> str:
        return self.HELP_MESSAGE

    @classmethod
    def help_short(cls) -> str:
        return cls.SHORT_HELP_MESSAGE

    def _separate_key_and_value(self) -> List:
        key_and_value = []
        for arg in self._only_args_starting_with_double_dash(self.args):
            split = arg.split('=')
            key_and_value.append((self._remove_double_dash(split[0]), split[1]))
        return key_and_value

    def _append_key_and_value_to_config(self, config: Settings, key_and_value: List):
        for key_value in key_and_value:
            try:
                config.set(key_value[0], key_value[1])
            except KeyError:
                print(f'Cannot set \"{key_value[0]}\", not valid configuration.\n')
                exit(1)

    def _only_args_starting_with_double_dash(self, args: List) -> List[str]:
        return [arg for arg in args if arg.startswith('--')]

    def _remove_double_dash(self, key: str) -> str:
        return key.replace('--', '')



from guet.commands.command import Command
from guet.config.committer import filter_committers_with_initials, Committer
from guet.config.get_committers import get_committers
from guet.config.set_author import set_committer_as_author
from guet.config.set_current_committers import set_current_committers as set_committers


class SetCommittersCommand(Command):

    def execute_hook(self) -> None:
        committers = get_committers()
        committer_initials = self.args
        committers_to_set = filter_committers_with_initials(committers, committer_initials)

        correct_number_of_committers_present = len(committers_to_set) is len(committer_initials)

        if not correct_number_of_committers_present:
            for initials in committer_initials:
                committer_with_initial_present = False
                for committer in committers:
                    if committer.initials == initials:
                        committer_with_initial_present = True
                if not committer_with_initial_present:
                    print(f"No committer exists with initials '{initials}'")
        else:
            set_committer_as_author(committers_to_set[0])
            set_committers(committers_to_set)

    def help(self):
        pass

    @classmethod
    def help_short(cls) -> str:
        return 'Set the current committers'



from os.path import join

from guet import constants
from guet.config import configuration_directory


def most_recent_committers_set():
    f = open(join(configuration_directory, constants.COMMITTERS_SET), 'r')
    line = f.readline()
    f.close()
    split = line.rstrip().split(',')
    return int(split[len(split)-1])




import traceback

from guet.config.set_errors import set_errors


def log_on_error(f):
    def wrapper():
        try:
            f()
        except Exception:
            print('An error has occurred, please refer to error logs for more information\n')
            stack_tract = traceback.format_exc()
            set_errors(stack_tract)
            exit(1)

    return wrapper



import subprocess


def configure_git_author(name: str, email: str) -> None:
    process = subprocess.Popen(['git', 'config', 'user.name', name])
    process.wait()
    process = subprocess.Popen(['git', 'config', 'user.email', email])
    process.wait()



from os.path import join

from guet import constants
from guet.config import configuration_directory
from guet.files.write_lines import write_lines
from guet.settings.settings import Settings


def set_config(settings: Settings) -> None:
    lines = settings.write()
    write_lines(join(configuration_directory, constants.CONFIG), lines)



from guet.commands import HelpCommand
from guet.commands.command import Command


class CommandFactory:
    def __init__(self, command_builder_map):
        self.command_builder_map = command_builder_map

    def create(self, args: list) -> Command:
        if len(args) > 0:
            command_arg = args[0]
            initialized_command = self.command_builder_map[command_arg](args)
            return initialized_command
        else:
            return HelpCommand(args, self.command_builder_map)



from guet.git.hook_present import hook_present


def any_hooks_present(git_path: str):
    pre_commit_present = hook_present(git_path, 'pre-commit')
    post_commit_present = hook_present(git_path, 'post-commit')
    commit_msg_present = hook_present(git_path, 'commit-msg')

    return pre_commit_present or post_commit_present or commit_msg_present



from typing import List

from guet.git.set_author import configure_git_author

from guet.config.committer import Committer
from guet.config.get_current_committers import get_current_committers
from guet.config.set_author import set_committer_as_author
from guet.config.set_current_committers import set_current_committers
from guet.util.errors import log_on_error


@log_on_error
def post_commit():
    committers = _rotate_fist_commiter_to_last_committer(get_current_committers())
    set_committer_as_author(committers[0])
    set_current_committers(committers)
    configure_git_author(committers[0].name, committers[0].email)


def _rotate_fist_commiter_to_last_committer(committers: List[Committer]):
    new_last_committer = committers.pop(0)
    committers.append(new_last_committer)
    return committers



from os.path import join
from typing import List

from guet import constants
from guet.config import configuration_directory
from guet.config.committer import Committer, filter_committers_with_initials
from guet.config.get_committers import get_committers


def get_current_committers() -> List[Committer]:
    f = open(join(configuration_directory, constants.COMMITTERS_SET), 'r')
    line = f.readline()
    f.close()
    committer_initials = line.rstrip().split(',')
    del committer_initials[-1]
    committers = get_committers()
    committers_with_initials = filter_committers_with_initials(committers, committer_initials)
    final = []
    for initials in committer_initials:
        for committer in committers_with_initials:
            if committer.initials == initials:
                final.append(committer)
    return final





from enum import Enum


def _append_nothing(base_name: str):
    return base_name


def _append_guet(base_name: str):
    return base_name + '-guet'


def _should_not_be_called(base_name: str):
    return RuntimeError


class HookMode(Enum):
    NEW_OR_OVERWRITE = _append_nothing
    CREATE_ALONGSIDE = _append_guet
    CANCEL = _should_not_be_called



"""
Copyright 2018 Christopher M. Boyer

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List


class Command:
    @classmethod
    def get_short_help_message(cls):
        return cls.help_short()

    def __init__(self, args: List[str], args_needed: bool = True):
        self.args_needed = args_needed
        self.args = args[1:]

    def execute(self) -> None:
        no_args_given = len(self.args) == 0
        if no_args_given and self.args_needed:
            self._print_help_message()
        else:
            self.execute_hook()

    def execute_hook(self) -> None:
        raise NotImplementedError

    def help(self) -> str:
        raise NotImplementedError

    @classmethod
    def help_short(cls) -> str:
        raise NotImplementedError

    def _print_help_message(self) -> None:
        print(f'{self.help()}\n')



from guet.hooks.commit_msg import commit_msg as _commit_msg
from guet.hooks.post_commit import post_commit as _post_commit
from guet.hooks.pre_commit import pre_commit as _pre_commit


def manage(hook: str):
    if hook.endswith('commit-msg'):
        _commit_msg()
    elif hook.endswith('pre-commit'):
        _pre_commit()
    elif hook.endswith('post-commit'):
        _post_commit()



from os.path import join
from typing import List

from guet import constants
from guet.config import configuration_directory
from guet.config.committer import Committer

_INITIALS_POSITION = 0
_NAME_POSITION = 1
_EMAIL_POSITION = 2


def get_committers() -> List[Committer]:
    committers_file = open(join(configuration_directory, constants.COMMITTERS), 'r')
    committers_lines = committers_file.readlines()
    committers_file.close()
    return [_convert_committer_line_to_committer(line) for line in committers_lines]


def _convert_committer_line_to_committer(committer_line: str) -> Committer:
    split = committer_line.rstrip().split(',')
    return Committer(split[_NAME_POSITION], split[_EMAIL_POSITION], split[_INITIALS_POSITION])



from guet.config.most_recent_committers_set import most_recent_committers_set
from guet.currentmillis import current_millis
from guet.util.errors import log_on_error


@log_on_error
def pre_commit():
    now = current_millis()
    twenty_four_hours = 86400000
    twenty_four_hours_ago = now - twenty_four_hours
    set_time = most_recent_committers_set()
    if set_time < twenty_four_hours_ago:
        print("\nYou have not reset pairs in over twenty four hours!\nPlease reset your pairs by using guet set and including your pairs' initials\n")
        exit(1)
    else:
        exit(0)




from enum import Enum
from os import chmod, stat
from os.path import join
from guet.git._hook_mode import HookMode

HOOK_FILE_LINES = [
    '#! /usr/bin/env python3\n',
    'from guet.hooks import manage\n',
    'import sys\n',
    'manage(sys.argv[0])\n',
]


class Hooks(Enum):
    PRE_COMMIT = 'pre-commit'
    COMMIT_MSG = 'commit-msg'
    POST_COMMIT = 'post-commit'


def create_hook(hook_folder_path: str, hook: Hooks, create_mode: HookMode):
    final_hook_path = join(hook_folder_path, create_mode(hook.value))
    f = open(final_hook_path, 'w')
    f.writelines(HOOK_FILE_LINES)
    f.close()

    st = stat(final_hook_path)
    chmod(final_hook_path, st.st_mode | 0o111)



from guet.config.already_initialized import already_initialized
from guet.config.initialize import initialize
from .command import Command


class InitDataSourceCommand(Command):

    def __init__(self, args):
        super().__init__(args, args_needed=False)

    def execute_hook(self):
        extra_arguments_given = len(self.args) > 0
        if extra_arguments_given:
            print('Invalid arguments.\n\n   {}'.format(self.help()))
        else:
            if not already_initialized():
                initialize()
            else:
                print('Config folder already exists.')

    def help(self):
        return ''

    @classmethod
    def help_short(cls) -> str:
        return 'Initialize guet for use'



from guet.commands.command import Command
from guet.git.any_hooks_present import any_hooks_present
from guet.git.create_hook import create_hook, HookMode, Hooks
from guet.git.git_path_from_cwd import git_hook_path_from_cwd
from guet.git.git_present_in_cwd import git_present_in_cwd


class StartCommand(Command):

    def __init__(self, args):
        super().__init__(args, args_needed=False)

    def execute_hook(self) -> None:
        if git_present_in_cwd():
            hook_path = git_hook_path_from_cwd()
            if not any_hooks_present(hook_path):
                self._create_all_hooks(hook_path, HookMode.NEW_OR_OVERWRITE)
            else:
                hook_mode = None
                print(
                    'There is already commit hooks in this project. Would you like to overwrite (o), create (c) the file and put it in the hooks folder, or cancel (x)?')
                val = input()
                if val == 'o':
                    hook_mode = HookMode.NEW_OR_OVERWRITE
                elif val == 'c':
                    hook_mode = HookMode.CREATE_ALONGSIDE
                else:
                    hook_mode = HookMode.CANCEL
                self._create_all_hooks(hook_path, hook_mode)

        else:
            print('Git not initialized in this directory.')

    def _create_all_hooks(self, hook_folder_path: str, hook_mode: HookMode):
        if hook_mode is not HookMode.CANCEL:
            create_hook(hook_folder_path, Hooks.PRE_COMMIT, hook_mode)
            create_hook(hook_folder_path, Hooks.POST_COMMIT, hook_mode)
            create_hook(hook_folder_path, Hooks.COMMIT_MSG, hook_mode)

    def help(self):
        pass

    @classmethod
    def help_short(cls) -> str:
        return 'Start guet usage in the repository at current directory'



import unittest

from guet.commands import HelpCommand
from guet.commands.command import Command
from guet.factory import CommandFactory


class MockCommand(Command):
    def execute_hook(self) -> None:
        pass

    def help(self) -> str:
        pass

    @classmethod
    def help_short(cls) -> str:
        pass


class NotMockCommand(Command):
    def execute_hook(self) -> None:
        pass

    def help(self) -> str:
        pass

    @classmethod
    def help_short(cls) -> str:
        pass


class TestCommandFactory(unittest.TestCase):

    def test_returns_command_from_builder_method_that_matches(self):
        builder_map = dict()
        builder_map['command'] = MockCommand
        builder_map['not-command'] = NotMockCommand

        command_factory = CommandFactory(builder_map)
        args = ['command']
        result = command_factory.create(args)
        self.assertEqual(MockCommand, type(result))

    def test_returns_help_command_if_no_command_provided(self):
        command_factory = CommandFactory(dict())
        result = command_factory.create([])
        self.assertEqual(HelpCommand, type(result))

    def test_help_command_has_command_builder_map(self):
        command_builder_map = dict()
        command_builder_map['command'] = lambda args: Command(args)
        command_factory = CommandFactory(command_builder_map)
        result = command_factory.create([])
        self.assertEqual(command_builder_map, result.command_builder_map)



from guet.commands.command import Command


class HelpCommand(Command):
    _REQUIRED_ARGS_IN_CORRECT_ORDER = []

    def __init__(self, args, command_builder_map=None):
        super().__init__(args)
        if command_builder_map is None:
            command_builder_map = dict()
        self.command_builder_map = command_builder_map

    def help(self):
        help_message = 'usage: guet <command>\n'
        for key in self.command_builder_map:
            short_help_message = self.command_builder_map[key].get_short_help_message()
            help_message += '\n   {} -- {}'.format(key, short_help_message)
        return help_message + '\n'

    def execute_hook(self) -> None:
        pass

    @classmethod
    def help_short(cls) -> str:
        pass



from os.path import join
from typing import List

from guet import constants
from guet.config import configuration_directory
from guet.files.write_lines import write_lines


def set_errors(error_lines: List[str]):
    write_lines(join(configuration_directory, constants.ERRORS), error_lines)



from guet.config.add_committer import add_committer
from guet.config.already_initialized import already_initialized
from .command import Command


class AddUserCommand(Command):

    def __init__(self, args):
        super().__init__(args)

    def execute_hook(self):
        if len(self.args) != 3:
            if len(self.args) > 3:
                print('Too many arguments.')
            else:
                print('Not enough arguments.')
                print('')
                print(self.help())
                print('')
        else:
            if already_initialized():
                add_committer(self.args[0], self.args[1], self.args[2])
            else:
                print('guet has not been initialized yet! Please do so by running the command "guet init".')

    def help(self):
        return 'usage: guet add <initials> <"name"> <email>'

    @classmethod
    def help_short(cls) -> str:
        return 'Add committer to the list of available committers'



import time
from os.path import join
from typing import List

from guet import constants
from guet.config import configuration_directory
from guet.config.committer import Committer


def set_current_committers(committers: List[Committer]) -> None:
    committers_set_file = open(join(configuration_directory, constants.COMMITTERS_SET), 'w')
    committers_set_file.write(_format_committers_to_committers_set_format(committers))
    committers_set_file.close()


def _format_committers_to_committers_set_format(committers: List[Committer]) -> str:
    current_time_in_millis = int(round(time.time() * 1000))
    return ','.join([committer.initials for committer in committers]) + f',{current_time_in_millis}\n'





import sys
import traceback

from guet.commands import AddUserCommand, InitDataSourceCommand, SetCommittersCommand, StartCommand
from guet.commands.config import ConfigSetCommand
from guet.factory import CommandFactory
from guet.util.errors import log_on_error


def _command_builder_map():
    command_builder_map = dict()
    command_builder_map['add'] = AddUserCommand
    command_builder_map['init'] = InitDataSourceCommand
    command_builder_map['set'] = SetCommittersCommand
    command_builder_map['start'] = StartCommand
    command_builder_map['config'] = ConfigSetCommand
    return command_builder_map


@log_on_error
def main():
    command_factory = CommandFactory(_command_builder_map())
    command = command_factory.create(sys.argv[1:])
    command.execute()



from os.path import join
from typing import List


def edit_commit_msg(git_path: str, message: List[str]) -> None:
    f = open(join(git_path, 'COMMIT_EDITMSG'), 'w')
    f.writelines(message)
    f.close()



from os.path import join

from guet import constants
from guet.config import configuration_directory
from guet.config.committer import Committer


def set_committer_as_author(committer: Committer):
    f = open(join(configuration_directory, constants.AUTHOR_NAME), 'w')
    f.write(f'{committer.name}\n')
    f.close()
    f = open(join(configuration_directory, constants.AUTHOR_EMAIL), 'w')
    f.write(f'{committer.email}\n')
    f.close()
